/* $Id: mesh.c,v 1.13 1998/02/23 06:21:02 svmcguir Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "util.h"

void meshDataAllocate(meshData *md){
  /* Allocate memory for md.valueIfHeld and md.value
   */

  allocate2dDouble(&(md->valueIfHeld),md->option.pData.times,md->option.pData.paths);
  allocate2dDouble(&(md->newValueIfHeld),md->option.pData.times,md->option.pData.paths);
  allocate2dDouble(&(md->value),md->option.pData.times,md->option.pData.paths);
  allocate2dDouble(&(md->newValue),md->option.pData.times,md->option.pData.paths);

  allocate3dDouble(&(md->weights), md->option.pData.times,md->option.pData.paths,
                                   md->meshPaths );
}

void meshSetFinalValue(meshData *md){
  /* Set the values on the final timeslice to the payoffs or 0, whichever
   * is greater.
   */

  int path,time;

  time = md->option.pData.times-1;

  for(path=0;path<md->option.pData.paths;path++){
    md->value[time][path] = dMax(md->option.payoff[time][path],0);
    md->valueIfHeld[time][path] = 0;
    md->newValue[time][path] = md->value[time][path];
    md->newValueIfHeld[time][path] = md->valueIfHeld[time][path];
  }
}

void meshDataFill(meshData *md){
  /* Working backwards from the next to last timeslice (last is handled 
   * by meshSetFinalValue) calculate the values of the option if it is held
   * (equation 6).  Then set the value of the option to the greater of the 
   * value if held and the payoff.  The first timeslice is left for 
   * meshMeshEstimate().
   */

  int time, path, fromPath, toPath, basePath,underlier;
  static double **tp, *sum;
  static int allocated = -1;
  double clusterProb;

  if(allocated == -1){
    allocate2dDouble(&tp, md->option.pData.paths, md->meshPaths);
    allocate1dDouble(&sum, md->meshPaths);
    allocated = 1;
  }

  for(time=md->option.pData.times-2; time>=0; time--){
    /* Get all the transition probabilities at once */
    for(fromPath=0; fromPath < md->option.pData.paths; fromPath++){
      for(toPath=0; toPath < md->meshPaths; toPath++){
	tp[fromPath][toPath] = pathTransitionProb(&(md->option.pData), time, fromPath, toPath);
      }
    }

    /* The denominators of the weights are sums of tp[fromPath][toPath] over fromPath. */
    for(toPath=0; toPath<md->meshPaths; toPath++){
      sum[toPath] = 0;
      for(fromPath=0; fromPath<md->meshPaths; fromPath+=md->option.pData.clusterSize){
	sum[toPath] += tp[fromPath][toPath];
	clusterProb=1;
	for(underlier=0;underlier<md->option.pData.underliers;underlier++){
	  clusterProb *=  (1/sqrt(md->option.pData.clusterVariance*2*M_PI))*
	    exp(-pow(md->option.pData.price[time+1][toPath][underlier]-md->option.pData.price[time+1][fromPath][underlier],2)/
		(2*md->option.pData.clusterVariance));                       
	}
	sum[toPath] += (md->option.pData.clusterSize-1)*clusterProb;

      }
    }
    
    /* Now, weight[fromPath][toPath] = tp[fromPath][toPath]/sum[toPath]
     */
    for(fromPath=0; fromPath<md->option.pData.paths; fromPath++){
      md->valueIfHeld[time][fromPath] = 0;
      for(toPath=0; toPath<md->meshPaths; toPath++){
	/* Store value of weights */
	  md->weights[time][fromPath][toPath]=tp[fromPath][toPath]/sum[toPath];
/*
	  printf("weights[%i][%i][%i] = %lf\n",time,fromPath,toPath,md->weights[time][fromPath][toPath]); 
*/
      
	/* Compute value if held */
	md->valueIfHeld[time][fromPath] +=md->weights[time][fromPath][toPath]*
	  md->value[time+1][toPath];
      }
      md->value[time][fromPath] = dMax(md->valueIfHeld[time][fromPath], md->option.payoff[time][fromPath]);
    }
  }

    /* Now do [0][0] */  
/*   for(fromPath=0; fromPath < md->option.pData.paths; fromPath++){ */
/*     for(toPath=0; toPath < md->meshPaths; toPath++){ */
/*       tp[fromPath][toPath] = pathTransitionProb(&(md->option.pData), 0, fromPath, toPath); */
/*     } */
/*   } */
  
   /* The denominators of the weights are sums of tp[fromPath][toPath] over fromPath. */ 
/*   for(toPath=0; toPath<md->meshPaths; toPath++){ */
/*     sum[toPath] = 0; */
/*     for(fromPath=0; fromPath<md->meshPaths; fromPath++){ */
/*       sum[toPath] += tp[fromPath][toPath]; */
/*     } */
/*   } */
/*   md->valueIfHeld[0][0] = 0;  */
/*   for(toPath=0; toPath<md->meshPaths; toPath++){  */
/*     md->valueIfHeld[0][0] += tp[0][toPath]/sum[toPath]*md->value[1][toPath];  */
/*   }  */
/*   md->value[0][0] = dMax(md->valueIfHeld[0][fromPath], md->option.payoff[0][fromPath]);  */

}

double meshPathEstimate(meshData *md){
  int time, path;
  double value;

  if(md->option.payoff[0][0] >= md->valueIfHeld[0][0]){
    return(md->option.payoff[0][0]);
  }

  value = 0;
  for(path=md->meshPaths; path < md->option.pData.paths; path++){
    for(time=1; time < md->option.pData.times; time++){
      if(md->option.payoff[time][path] > md->valueIfHeld[time][path]){
	value += md->option.payoff[time][path];
	break;
      }
    }
  }

  return(value/(double)(md->option.pData.paths-md->meshPaths));
}

